/*
 *  $Id: psf.c 21201 2018-06-24 19:34:11Z yeti-dn $
 *  Copyright (C) 2017-2018 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libprocess/gwyprocesstypes.h>
#include <libprocess/arithmetic.h>
#include <libprocess/grains.h>
#include <libprocess/inttrans.h>
#include <libprocess/stats.h>
#include <libgwydgets/gwydgetutils.h>
#include <libgwydgets/gwycombobox.h>
#include <libgwydgets/gwycheckboxes.h>
#include <libgwydgets/gwystock.h>
#include <libgwymodule/gwymodule-process.h>
#include <app/gwymoduleutils.h>
#include <app/gwyapp.h>
#include "mfmops.h"
#include "preview.h"

#define PSF_RUN_MODES GWY_RUN_INTERACTIVE

#define field_convolve_default(field, kernel) \
    gwy_data_field_area_ext_convolve((field), \
                                     0, 0, \
                                     gwy_data_field_get_xres(field), \
                                     gwy_data_field_get_yres(field), \
                                     (field), (kernel), \
                                     GWY_EXTERIOR_MIRROR_EXTEND, 0.0, TRUE)

typedef enum {
    GWY_PSF_DISPLAY_DATA       = 0,
    GWY_PSF_DISPLAY_PSF     = 1,
    GWY_PSF_DISPLAY_CONVOLVED  = 2,
    GWY_PSF_DISPLAY_DIFFERENCE = 3
} GwyPSFDisplayType;

typedef enum {
    PSF_OUTPUT_PSF       = (1 << 0),
    PSF_OUTPUT_CONVOLVED = (1 << 1),
    PSF_OUTPUT_DIFFERECE = (1 << 2),
    PSF_OUTPUTS_MASK     = (1 << 3) - 1,
} PSFOutputType;

typedef struct {
    gdouble sigma;
    GwyAppDataId op1;
    GwyAppDataId op2;
    GwyPSFDisplayType display;
    PSFOutputType output_type;
} PSFArgs;

typedef struct {
    PSFArgs *args;
    GtkObject *sigma;
    GtkWidget *chooser_op2;
    GtkWidget *display;
    GSList *output_type;
    GwyDataField *result;
    GwyDataField *dfield;
    GtkWidget *view;
    GwyContainer *mydata;
    GtkWidget *dialog;
    gboolean in_update;
    GwySIValueFormat *vf;
    GtkWidget *tf_width;
} PSFControls;

static gboolean module_register    (void);
static void     psf                (GwyContainer *data,
                                    GwyRunType run);
static gboolean psf_dialog         (PSFArgs *args,
                                    GwyContainer *data);
static void     psf_load_args      (GwyContainer *container,
                                    PSFArgs *args);
static void     psf_save_args      (GwyContainer *container,
                                    PSFArgs *args);
static void     psf_sanitize_args  (PSFArgs *args);
static void     sigma_changed      (GtkAdjustment *adj,
                                    PSFControls *controls);
static void     psf_data_changed   (GwyDataChooser *chooser,
                                    PSFControls *controls);
static gboolean psf_data_filter    (GwyContainer *data,
                                    gint id,
                                    gpointer user_data);
static void     preview            (PSFControls *controls,
                                    PSFArgs *args);
static void     display_changed    (GtkComboBox *combo,
                                    PSFControls *controls);
static void     output_type_changed(GtkToggleButton *toggle,
                                    PSFControls *controls);
static void     update_sensitivity (PSFControls *controls);
static void     calculate_tf       (GwyDataField *measured,
                                    GwyDataField *ideal,
                                    GwyDataField *psf,
                                    gdouble sigma);
static gdouble  measure_tf_width   (GwyDataField *psf);
static void     estimate_sigma     (PSFArgs *args);
static void     create_output_field(GwyDataField *dfield,
                                    GwyAppDataId dataid,
                                    const gchar *name);

static GwyAppDataId op2_id = GWY_APP_DATA_ID_NONE;

static const PSFArgs psf_defaults = {
    10, GWY_APP_DATA_ID_NONE, GWY_APP_DATA_ID_NONE,
    GWY_PSF_DISPLAY_PSF, PSF_OUTPUT_PSF,
};

static const GwyEnum output_types[] = {
    { N_("PSF"),        PSF_OUTPUT_PSF,       },
    { N_("Convolved"),  PSF_OUTPUT_CONVOLVED, },
    { N_("Difference"), PSF_OUTPUT_DIFFERECE, },
};

enum {
    OUTPUT_NTYPES = G_N_ELEMENTS(output_types)
};

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("Transfer function estimation"),
    "Petr Klapetek <klapetek@gwyddion.net>",
    "2.0",
    "David Nečas (Yeti) & Petr Klapetek",
    "2017",
};

GWY_MODULE_QUERY2(module_info, psf)

static gboolean
module_register(void)
{
    gwy_process_func_register("psf",
                              (GwyProcessFunc)&psf,
                              N_("/_Statistics/_Transfer Function Guess..."),
                              NULL,
                              PSF_RUN_MODES,
                              GWY_MENU_FLAG_DATA,
                              N_("Estimate transfer function from known data and image"));

    return TRUE;
}

static void
psf(GwyContainer *data, GwyRunType run)
{
    GwyDataField *dfield1, *dfield2, *psf, *convolved = NULL, *difference;
    PSFArgs args;
    gboolean ok;
    GQuark quark;
    GwyContainer *mydata;

    g_return_if_fail(run & PSF_RUN_MODES);

    psf_load_args(gwy_app_settings_get(), &args);

    gwy_app_data_browser_get_current(GWY_APP_DATA_FIELD_ID, &args.op1.id,
                                     GWY_APP_CONTAINER_ID, &args.op1.datano,
                                     0);

    ok = psf_dialog(&args, data);
    psf_save_args(gwy_app_settings_get(), &args);
    if (!ok)
        return;

    mydata = gwy_app_data_browser_get(args.op1.datano);
    quark = gwy_app_get_data_key_for_id(args.op1.id);
    dfield1 = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    mydata = gwy_app_data_browser_get(args.op2.datano);
    quark = gwy_app_get_data_key_for_id(args.op2.id);
    dfield2 = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    psf = gwy_data_field_new_alike(dfield1, TRUE);
    calculate_tf(dfield1, dfield2, psf, args.sigma/100.0);
    if (args.output_type & (PSF_OUTPUT_CONVOLVED | PSF_OUTPUT_DIFFERECE)) {
        convolved = gwy_data_field_new_alike(dfield1, TRUE);
        gwy_data_field_copy(dfield2, convolved, TRUE);
        field_convolve_default(convolved, psf);
        gwy_data_field_add(convolved, gwy_data_field_get_avg(dfield1));
    }

    if (args.output_type & PSF_OUTPUT_PSF)
        create_output_field(psf, args.op1, _("Transfer function"));
    if (args.output_type & PSF_OUTPUT_CONVOLVED)
        create_output_field(convolved, args.op1, _("Convolved"));
    if (args.output_type & PSF_OUTPUT_DIFFERECE) {
        difference = gwy_data_field_new_alike(dfield1, FALSE);
        gwy_data_field_subtract_fields(difference, convolved, dfield1);
        create_output_field(difference, args.op1, _("Difference"));
        g_object_unref(difference);
    }

    g_object_unref(psf);
    GWY_OBJECT_UNREF(convolved);
}

static gboolean
psf_dialog(PSFArgs *args, GwyContainer *data)
{
    static const GwyEnum psf_displays[] = {
        { N_("Data"),              GWY_PSF_DISPLAY_DATA,       },
        { N_("Transfer function"), GWY_PSF_DISPLAY_PSF,        },
        { N_("Convolved"),         GWY_PSF_DISPLAY_CONVOLVED,  },
        { N_("Difference"),        GWY_PSF_DISPLAY_DIFFERENCE, },
    };

    GtkWidget *dialog, *table, *hbox, *spin, *label;
    GwyDataChooser *chooser;
    PSFControls controls;
    gint response, row;
    GQuark quark;
    GwyContainer *mydata;

    controls.args = args;
    controls.in_update = TRUE;

    dialog = gtk_dialog_new_with_buttons(_("Estimate Transfer Function"), NULL, 0,
                                         _("_Fit sigma"), RESPONSE_ESTIMATE,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
    controls.dialog = dialog;

    hbox = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(hbox),
                       FALSE, FALSE, 4);

    mydata = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    controls.dfield = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));
    controls.result = gwy_data_field_new_alike(controls.dfield, TRUE);

    controls.mydata = gwy_container_new();
    gwy_container_set_object_by_name(controls.mydata, "/0/data",
                                     controls.result);
    gwy_app_sync_data_items(data, controls.mydata, args->op1.id, 0, FALSE,
                            GWY_DATA_ITEM_PALETTE,
                            GWY_DATA_ITEM_MASK_COLOR,
                            GWY_DATA_ITEM_RANGE,
                            GWY_DATA_ITEM_REAL_SQUARE,
                            0);
    controls.view = create_preview(controls.mydata, 0, PREVIEW_SIZE, TRUE);
    gtk_box_pack_start(GTK_BOX(hbox), controls.view, FALSE, FALSE, 4);

    table = gtk_table_new(9 + OUTPUT_NTYPES, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 6);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 4);
    row = 0;

    controls.chooser_op2 = gwy_data_chooser_new_channels();
    chooser = GWY_DATA_CHOOSER(controls.chooser_op2);
    gwy_data_chooser_set_active_id(chooser, &args->op2);
    gwy_data_chooser_set_filter(chooser, psf_data_filter, &args->op1, NULL);
    g_signal_connect(chooser, "changed",
                     G_CALLBACK(psf_data_changed), &controls);
    gwy_table_attach_adjbar(table, row, _("_Ideal response:"), NULL,
                            GTK_OBJECT(chooser), GWY_HSCALE_WIDGET_NO_EXPAND);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;

    controls.display
        = gwy_enum_combo_box_new(psf_displays, G_N_ELEMENTS(psf_displays),
                                 G_CALLBACK(display_changed),
                                 &controls, args->display, TRUE);
    gwy_table_attach_adjbar(table, row, _("_Display:"), NULL,
                            GTK_OBJECT(controls.display),
                            GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;

    controls.sigma = gtk_adjustment_new(args->sigma, 1e-5, 1e2, 1e-5, 1, 0);
    spin = gwy_table_attach_adjbar(table, row, _("_Sigma:"), "%",
                                   controls.sigma, GWY_HSCALE_LOG);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 7);
    g_object_set_data(G_OBJECT(controls.sigma), "controls", &controls);
    g_signal_connect(controls.sigma, "value-changed",
                     G_CALLBACK(sigma_changed), &controls);
    row++;

    controls.vf = gwy_data_field_get_value_format_xy(controls.dfield,
                                                     GWY_SI_UNIT_FORMAT_VFMARKUP,
                                                     NULL);
    /* The width is usually rather small. */
    controls.vf->precision++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    label = gtk_label_new(_("TF width:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label,
                     0, 1, row, row+1, GTK_FILL, 0, 0, 0);

    label = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL(label), controls.vf->units);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label,
                     2, 3, row, row+1, GTK_FILL, 0, 0, 0);

    controls.tf_width = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(controls.tf_width), 1.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), controls.tf_width,
                     1, 2, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    label = gtk_label_new(_("Output type:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 0, 2, row, row+1,
                     GTK_FILL, 0, 0, 0);
    row++;

    controls.output_type
        = gwy_check_boxes_create(output_types, OUTPUT_NTYPES,
                                 G_CALLBACK(output_type_changed), &controls,
                                 args->output_type);
    row = gwy_check_boxes_attach_to_table(controls.output_type,
                                          GTK_TABLE(table), 2, row);

    controls.in_update = FALSE;
    update_sensitivity(&controls);
    psf_data_changed(chooser, &controls);

    gtk_widget_show_all(dialog);
    do {
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            gtk_widget_destroy(dialog);
            g_object_unref(controls.mydata);
            gwy_si_unit_value_format_free(controls.vf);
            case GTK_RESPONSE_NONE:
            return FALSE;
            break;

            case GTK_RESPONSE_OK:
            break;

            case RESPONSE_PREVIEW:
            preview(&controls, args);
            break;

            case RESPONSE_ESTIMATE:
            gwy_app_wait_cursor_start(GTK_WINDOW(dialog));
            estimate_sigma(args);
            gwy_app_wait_cursor_finish(GTK_WINDOW(dialog));
            gtk_adjustment_set_value(GTK_ADJUSTMENT(controls.sigma),
                                     args->sigma);
            preview(&controls, args);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    } while (response != GTK_RESPONSE_OK);

    gtk_widget_destroy(dialog);
    g_object_unref(controls.mydata);
    gwy_si_unit_value_format_free(controls.vf);

    return TRUE;
}

static void
display_changed(GtkComboBox *combo, PSFControls *controls)
{
    controls->args->display = gwy_enum_combo_box_get_active(combo);
    preview(controls, controls->args);
}

static void
sigma_changed(GtkAdjustment *adj, PSFControls *controls)
{
    controls->args->sigma = gtk_adjustment_get_value(adj);
    preview(controls, controls->args);
}

static void
output_type_changed(G_GNUC_UNUSED GtkToggleButton *toggle,
                    PSFControls *controls)
{
    PSFArgs *args = controls->args;

    args->output_type = gwy_check_boxes_get_selected(controls->output_type);
    update_sensitivity(controls);
}

static void
update_sensitivity(PSFControls *controls)
{
    PSFArgs *args = controls->args;

    gtk_dialog_set_response_sensitive(GTK_DIALOG(controls->dialog),
                                      GTK_RESPONSE_OK,
                                      args->op2.datano && args->output_type);
}

static void
estimate_sigma(PSFArgs *args)
{
    GwyDataField *measured, *ideal;
    GQuark quark;
    GwyContainer *mydata;

    mydata = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    measured = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    mydata = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    ideal = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    args->sigma = gwy_data_field_find_regularization_sigma_for_psf(measured,
                                                                   ideal);
    /* Convert NSR to GUI parameter convention */
    args->sigma = 100.0*sqrt(args->sigma);
}

static void
create_output_field(GwyDataField *dfield,
                    GwyAppDataId dataid,
                    const gchar *name)
{
    GwyContainer *data;
    gint newid;

    data = gwy_app_data_browser_get(dataid.datano);
    newid = gwy_app_data_browser_add_data_field(dfield, data, TRUE);
    gwy_app_sync_data_items(data, data, dataid.id, newid, FALSE,
                            GWY_DATA_ITEM_GRADIENT,
                            GWY_DATA_ITEM_MASK_COLOR,
                            0);
    gwy_app_set_data_field_title(data, newid, name);
    gwy_app_channel_log_add_proc(data, dataid.id, newid);
}

static void
preview(PSFControls *controls, PSFArgs *args)
{
    GwyDataField *dfield1, *dfield2, *psf, *convolved;
    GQuark quark;
    GwyContainer *mydata;
    gdouble realsigma, tf_width;
    gchar *s;

    mydata = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    dfield1 = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    if (args->op2.datano < 0 || args->op2.id < 0) {
        if (args->display == GWY_PSF_DISPLAY_DATA)
            gwy_data_field_copy(dfield1, controls->result, TRUE);
        else
            gwy_data_field_clear(controls->result);
        gwy_data_field_data_changed(controls->result);
        return;
    }

    mydata = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    dfield2 = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    psf = gwy_data_field_new_alike(dfield1, TRUE);

    realsigma = args->sigma/100.0;
    calculate_tf(dfield1, dfield2, psf, realsigma);
    tf_width = measure_tf_width(psf);

    s = g_strdup_printf("%.*f",
                        controls->vf->precision,
                        tf_width/controls->vf->magnitude);
    gtk_label_set_text(GTK_LABEL(controls->tf_width), s);
    g_free(s);

    if (args->display == GWY_PSF_DISPLAY_DATA) {
        gwy_data_field_copy(dfield1, controls->result, TRUE);
        gwy_data_field_data_changed(controls->result);
    }
    else if (args->display == GWY_PSF_DISPLAY_PSF) {
        gwy_data_field_copy(psf, controls->result, TRUE);
        gwy_data_field_data_changed(controls->result);
    }
    else {
        convolved = gwy_data_field_new_alike(dfield1, TRUE);
        gwy_data_field_copy(dfield2, convolved, TRUE);
        field_convolve_default(convolved, psf);
        gwy_data_field_add(convolved, gwy_data_field_get_avg(dfield1));
        if (args->display == GWY_PSF_DISPLAY_DIFFERENCE)
            gwy_data_field_subtract_fields(convolved, convolved, dfield1);
        gwy_data_field_copy(convolved, controls->result, TRUE);
        gwy_data_field_data_changed(controls->result);
        gwy_object_unref(convolved);
    }
}

static void
psf_data_changed(GwyDataChooser *chooser, PSFControls *controls)
{
    GwyAppDataId *object = &controls->args->op2;

    gwy_data_chooser_get_active_id(chooser, object);
    gwy_debug("data: %d %d", object->datano, object->id);
    update_sensitivity(controls);
    preview(controls, controls->args);
}

static gboolean
psf_data_filter(GwyContainer *data, gint id, gpointer user_data)
{

    GwyAppDataId *object = (GwyAppDataId*)user_data;
    GwyDataField *op1, *op2;
    GQuark quark;

    quark = gwy_app_get_data_key_for_id(id);
    op1 = GWY_DATA_FIELD(gwy_container_get_object(data, quark));

    data = gwy_app_data_browser_get(object->datano);
    quark = gwy_app_get_data_key_for_id(object->id);
    op2 = GWY_DATA_FIELD(gwy_container_get_object(data, quark));

    /* It does not make sense to crosscorrelate with itself */
    if (op1 == op2)
        return FALSE;

    return !gwy_data_field_check_compatibility(op1, op2,
                                               GWY_DATA_COMPATIBILITY_RES
                                               | GWY_DATA_COMPATIBILITY_REAL
                                               | GWY_DATA_COMPATIBILITY_LATERAL);
}

static void
calculate_tf(GwyDataField *measured, GwyDataField *ideal,
             GwyDataField *psf, gdouble sigma)
{
    gwy_data_field_deconvolve_regularized(measured, ideal, psf, sigma);
}

static gdouble
measure_tf_width(GwyDataField *psf)
{
    GwyDataField *buf;
    gint xres, yres, xc, yc, i, j, k;
    const gdouble *d, *b;
    gdouble sx, sy, s2, sw, w, dx, dy, x, y;

    xres = gwy_data_field_get_xres(psf);
    yres = gwy_data_field_get_yres(psf);
    dx = gwy_data_field_get_xmeasure(psf);
    dy = gwy_data_field_get_ymeasure(psf);
    xc = xres/2;
    yc = yres/2;

    buf = gwy_data_field_duplicate(psf);
    gwy_data_field_threshold(buf, 0.05*gwy_data_field_get_max(buf), 0.0, 1.0);
    b = gwy_data_field_get_data_const(buf);
    if (b[yc*xres + xc] == 0.0) {
        g_warning("Cannot find the central TF peak.  Improve me!");
        g_object_unref(buf);
        return 0.0;
    }

    gwy_data_field_grains_extract_grain(buf, xc, yc);
    gwy_data_field_grains_grow(buf, 0.5*log(xres*yres),
                               GWY_DISTANCE_TRANSFORM_EUCLIDEAN, FALSE);
    d = gwy_data_field_get_data_const(psf);
    b = gwy_data_field_get_data_const(buf);
    sx = sy = sw = 0.0;
    k = 0;
    for (i = 0; i < yres; i++) {
        y = i*dy;
        for (j = 0; j < xres; j++, k++) {
            if (b[k] <= 0.0)
                continue;

            x = j*dx;
            w = fabs(d[k]);
            sw += w;
            sx += x*w;
            sy += y*w;
        }
    }
    sx /= sw;
    sy /= sw;
    s2 = 0.0;
    k = 0;
    for (i = 0; i < yres; i++) {
        y = i*dy;
        for (j = 0; j < xres; j++, k++) {
            if (b[k] <= 0.0)
                continue;

            x = j*dx;
            w = fabs(d[k]);
            s2 += ((x - sx)*(x - sx) + (y - sy)*(y - sy))*w;
        }
    }

    g_object_unref(buf);

    return sqrt(s2/sw);
}

static const gchar display_key[]     = "/module/psf/display";
static const gchar output_type_key[] = "/module/psf/output_type";
static const gchar sigma_key[]       = "/module/psf/sigma";

static void
psf_sanitize_args(PSFArgs *args)
{
    gwy_app_data_id_verify_channel(&args->op2);
    args->output_type &= PSF_OUTPUTS_MASK;
    args->sigma = CLAMP(args->sigma, 1e-5, 1e2);
}

static void
psf_load_args(GwyContainer *container, PSFArgs *args)
{
    *args = psf_defaults;

    gwy_container_gis_double_by_name(container, sigma_key, &args->sigma);
    gwy_container_gis_enum_by_name(container, display_key, &args->display);
    gwy_container_gis_enum_by_name(container, output_type_key,
                                   &args->output_type);
    args->op2 = op2_id;

    psf_sanitize_args(args);
}

static void
psf_save_args(GwyContainer *container, PSFArgs *args)
{
    op2_id = args->op2;

    gwy_container_set_double_by_name(container, sigma_key, args->sigma);
    gwy_container_set_enum_by_name(container, display_key, args->display);
    gwy_container_set_enum_by_name(container, output_type_key,
                                   args->output_type);
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
