/*
 *  $Id: volume_mfmrecalc.c 21148 2018-06-14 10:15:39Z yeti-dn $
 *  Copyright (C) 2017-2018 David Necas (Yeti).
 *  E-mail: yeti@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libprocess/brick.h>
#include <libgwydgets/gwycombobox.h>
#include <libgwydgets/gwystock.h>
#include <libgwydgets/gwydgetutils.h>
#include <libgwydgets/gwyradiobuttons.h>
#include <libgwymodule/gwymodule-volume.h>
#include <app/gwyapp.h>
#include <app/gwymoduleutils.h>
#include "../process/mfmops.h" //mfmops should go to libprocess

#define VMFM_RECALC_RUN_MODES (GWY_RUN_IMMEDIATE | GWY_RUN_INTERACTIVE)

typedef enum  {
    SIGNAL_PHASE_DEG = 0,
    SIGNAL_PHASE_RAD = 1,
    SIGNAL_FREQUENCY = 2,
    SIGNAL_AMPLITUDE_V = 3,
    SIGNAL_AMPLITUDE_M = 4
} MfmRecalcSignal;

enum {
    RESPONSE_RESET   = 1,
};

typedef struct {
    MfmRecalcSignal signal;
    gdouble spring_constant;
    gdouble quality;
    gdouble base_frequency;
    gdouble base_amplitude;

} VMfmRecalcArgs;

typedef struct {
    VMfmRecalcArgs *args;
    GSList *signal;
    GtkObject *spring_constant;
    GtkObject *quality;
    GtkWidget *quality_widget;
    GtkObject *base_frequency;
    GtkWidget *base_frequency_widget;
    GtkObject *base_amplitude;
    GtkWidget *base_amplitude_widget;

} VMfmRecalcControls;

static gboolean module_register                (void);
static void     volume_mfmrecalc               (GwyContainer *data,
                                                GwyRunType run);
static gboolean vmfm_recalc_dialog             (VMfmRecalcArgs *args,
                                                MfmRecalcSignal guess);
static void     update_sensitivity             (VMfmRecalcControls *controls);
static void     vmfm_recalc_sanitize_args      (VMfmRecalcArgs *args);
static void     vmfm_recalc_load_args          (GwyContainer *container,
                                                VMfmRecalcArgs *args);
static void     vmfm_recalc_save_args          (GwyContainer *container,
                                                VMfmRecalcArgs *args);
static void     signal_changed                 (GtkToggleButton *toggle,
                                                VMfmRecalcControls *controls);
static void     vmfm_recalc_dialog_update      (VMfmRecalcControls *controls);
static void     mfm_recalc_dialog_update_values(VMfmRecalcControls *controls,
                                                VMfmRecalcArgs *args);

static const VMfmRecalcArgs vmfm_recalc_defaults = {
    SIGNAL_PHASE_DEG, 40, 1000, 150, 0.2,
};

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("Converts the MFM data to force gradient."),
    "Petr Klapetek <klapetek@gwyddion.net>, Robb Puttock <robb.puttock@npl.co.uk>",
    "1.0",
    "David Nečas (Yeti) & Petr Klapetek",
    "2018",
};

GWY_MODULE_QUERY2(module_info, volume_mfmrecalc)

static gboolean
module_register(void)
{
    gwy_volume_func_register("volume_mfmrecalc",
                             (GwyVolumeFunc)&volume_mfmrecalc,
                             N_("/_Magnetic Data to Force Gradient..."),
                             NULL,
                             VMFM_RECALC_RUN_MODES,
                             GWY_MENU_FLAG_VOLUME,
                             N_("Recalculate to force gradient"));

    return TRUE;
}

static void
issue_warning(GtkWindow *window)
{
    GtkWidget *dialog;

    dialog = gtk_message_dialog_new(window,
                                    GTK_DIALOG_DESTROY_WITH_PARENT,
                                    GTK_MESSAGE_ERROR,
                                    GTK_BUTTONS_OK,
                                    _("Data value units must be "
                                      "deg, rad, m, Hz or V "
                                      "for the recalculation"));
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

static void
volume_mfmrecalc(GwyContainer *data, GwyRunType run)
{
    VMfmRecalcArgs args;
    GwyBrick *brick = NULL, *result;
    GwyDataField *preview;
    gint id, newid, xres, yres;
    gboolean ok;
    GwySIUnit *zunit;
    MfmRecalcSignal guess;
    GQuark quarks[2];
    const guchar *gradient;

    g_return_if_fail(run & VMFM_RECALC_RUN_MODES);
    vmfm_recalc_load_args(gwy_app_settings_get(), &args);
    gwy_app_data_browser_get_current(GWY_APP_BRICK, &brick,
                                     GWY_APP_BRICK_ID, &id,
                                     0);
    g_return_if_fail(GWY_IS_BRICK(brick));

    //guess the units
    zunit = gwy_brick_get_si_unit_w(brick);
    if (gwy_si_unit_equal_string(zunit, "deg") || gwy_si_unit_equal_string(zunit, "°"))
        guess = SIGNAL_PHASE_DEG;
    else if (gwy_si_unit_equal_string(zunit, "rad"))
        guess = SIGNAL_PHASE_RAD;
    else if (gwy_si_unit_equal_string(zunit, "Hz"))
        guess = SIGNAL_FREQUENCY;
    else if(gwy_si_unit_equal_string(zunit, "V"))
        guess = SIGNAL_AMPLITUDE_V;
    else if (gwy_si_unit_equal_string(zunit, "m"))
        guess = SIGNAL_AMPLITUDE_M;
    else {
        issue_warning(gwy_app_find_window_for_channel(data, id));
        return;
    }

    args.signal = guess;

    if (run == GWY_RUN_INTERACTIVE) {
        ok = vmfm_recalc_dialog(&args, guess);
        vmfm_recalc_save_args(gwy_app_settings_get(), &args);

        if (!ok)
            return;
    }

    result = gwy_brick_new_alike(brick, TRUE);

    if (args.signal == SIGNAL_AMPLITUDE_M) args.base_amplitude *= 1e-9;

    if (args.signal == SIGNAL_PHASE_DEG) {
        get_volume_force_gradient_from_phase(brick, result, args.spring_constant, args.quality);
        gwy_brick_multiply(result, M_PI/180);
    }
    else if (args.signal == SIGNAL_PHASE_RAD) {
        get_volume_force_gradient_from_phase(brick, result, args.spring_constant, args.quality);
    }
    else if (args.signal == SIGNAL_FREQUENCY) {
        get_volume_force_gradient_from_frequency_shift(brick, result, args.spring_constant, args.base_frequency);
    }
    else get_volume_force_gradient_from_amplitude_shift(brick, result, args.spring_constant, args.quality, args.base_amplitude);


    xres = gwy_brick_get_xres(result);
    yres = gwy_brick_get_yres(result);
    preview = gwy_data_field_new(xres, yres, xres, yres, FALSE);
    gwy_brick_mean_plane(result, preview, 0, 0, 0, xres, yres, -1, TRUE);

    newid = gwy_app_data_browser_add_brick(result, preview, data, TRUE);
    gwy_app_set_brick_title(data, id, _("Recalculated Data"));
    gwy_app_volume_log_add_volume(data, id, newid);

    quarks[0] = gwy_app_get_brick_palette_key_for_id(id);
    quarks[1] = gwy_app_get_brick_palette_key_for_id(newid);
    if (gwy_container_gis_string(data, quarks[0], &gradient))
        gwy_container_set_const_string(data, quarks[1], gradient);

}

static gboolean
vmfm_recalc_dialog(VMfmRecalcArgs *args, MfmRecalcSignal guess)
{

    GtkWidget *dialog, *table;
    VMfmRecalcControls controls;
    gint response, row;

    static const GwyEnum signals[] = {
    { N_("Phase (radians)"),   SIGNAL_PHASE_RAD, },
    { N_("Phase (degrees)"),   SIGNAL_PHASE_DEG, },
    { N_("Frequency shift"),   SIGNAL_FREQUENCY, },
    { N_("Amplitude (V)"),     SIGNAL_AMPLITUDE_V, },
    { N_("Amplitude (m)"),     SIGNAL_AMPLITUDE_M, },
     };


    controls.args = args;

    dialog = gtk_dialog_new_with_buttons(_("MFM Recalculate Data"),
                                         NULL, 0,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
    gwy_help_add_to_volume_dialog(GTK_DIALOG(dialog), GWY_HELP_DEFAULT);

    table = gtk_table_new(5, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_table_set_col_spacings(GTK_TABLE(table), 6);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), table, TRUE, TRUE, 4);
    row = 0;

    controls.signal
        = gwy_radio_buttons_create(signals, G_N_ELEMENTS(signals),
                                   G_CALLBACK(signal_changed),
                                   &controls, args->signal);
    row = gwy_radio_buttons_attach_to_table(controls.signal,
                                            GTK_TABLE(table), 2, row);
    row++;

    controls.spring_constant = gtk_adjustment_new(args->spring_constant,
                                                  0.01, 1000.0, 0.01, 1, 0);
    gwy_table_attach_adjbar(table, row, _("_Spring constant:"), "N/m",
                            controls.spring_constant, GWY_HSCALE_LOG);
    row++;

    controls.quality = gtk_adjustment_new(args->quality,
                                          0.01, 10000.0, 0.01, 1, 0);
    gwy_table_attach_adjbar(table, row, _("_Quality factor:"), NULL,
                            controls.quality, GWY_HSCALE_LOG);
    row++;

    controls.base_frequency = gtk_adjustment_new(args->base_frequency,
                                                 1, 1000000.0, 1, 10, 0);
    gwy_table_attach_adjbar(table, row, _("_Base frequency:"), "Hz",
                            controls.base_frequency, GWY_HSCALE_LOG);
    row++;

    controls.base_amplitude = gtk_adjustment_new(args->base_amplitude,
                                                 0.01, 1000.0, 0.01, 1, 0);
    gwy_table_attach_adjbar(table, row, _("_Base amplitude:"), "V, nm",
                            controls.base_amplitude, GWY_HSCALE_LOG);
    row++;

    update_sensitivity(&controls);

    gtk_widget_show_all(dialog);
    do {
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            gtk_widget_destroy(dialog);
            case GTK_RESPONSE_NONE:
            return FALSE;
            break;

            case GTK_RESPONSE_OK:
            mfm_recalc_dialog_update_values(&controls, args);
            break;

            case RESPONSE_RESET:
            *args = vmfm_recalc_defaults;
            args->signal = guess;
            vmfm_recalc_dialog_update(&controls);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    } while (response != GTK_RESPONSE_OK);

    gtk_widget_destroy(dialog);

    return TRUE;
}

static void
update_sensitivity(VMfmRecalcControls *controls)
{
    VMfmRecalcArgs *args = controls->args;
    GtkWidget *label;
    gboolean is_phase, is_freq, is_amplitude;

    is_phase = (args->signal == SIGNAL_PHASE_DEG
                || args->signal == SIGNAL_PHASE_RAD);
    is_freq = (args->signal == SIGNAL_FREQUENCY);
    is_amplitude = (args->signal == SIGNAL_AMPLITUDE_V
                    || args->signal == SIGNAL_AMPLITUDE_M);

    gwy_table_hscale_set_sensitive(controls->base_frequency, is_freq);
    gwy_table_hscale_set_sensitive(controls->quality, is_phase || is_amplitude);
    gwy_table_hscale_set_sensitive(controls->base_amplitude, is_amplitude);

    /* Reset all to insensitive and then make the enabled buttons sensitive.
     * When there are more choices for one quantity type, this may become more
     * complex. */
    gwy_radio_buttons_set_sensitive(controls->signal, FALSE);
    gtk_widget_set_sensitive(gwy_radio_buttons_find(controls->signal,
                                                    args->signal), TRUE);

    label = gwy_table_hscale_get_units(controls->base_amplitude);
    /* This is correct, the signal is in [m], but the user enters base
     * amplitude in [nm]. */
    if (args->signal == SIGNAL_AMPLITUDE_M)
        gtk_label_set_text(GTK_LABEL(label), "nm");
    else
        gtk_label_set_text(GTK_LABEL(label), "V");
}

static void
vmfm_recalc_dialog_update(VMfmRecalcControls *controls)
{
    VMfmRecalcArgs *args = controls->args;

    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->spring_constant),
                             args->spring_constant);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->quality),
                             args->quality);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->base_frequency),
                             args->base_frequency);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->base_amplitude),
                             args->base_amplitude);
    gwy_enum_combo_box_set_active(GTK_COMBO_BOX(controls->signal),
                                  args->signal);
}

static void
mfm_recalc_dialog_update_values(VMfmRecalcControls *controls,
                                VMfmRecalcArgs *args)
{
    args->spring_constant
        = gtk_adjustment_get_value(GTK_ADJUSTMENT(controls->spring_constant));
    args->quality
        = gtk_adjustment_get_value(GTK_ADJUSTMENT(controls->quality));
    args->base_frequency
        = gtk_adjustment_get_value(GTK_ADJUSTMENT(controls->base_frequency));
    args->base_amplitude
        = gtk_adjustment_get_value(GTK_ADJUSTMENT(controls->base_amplitude));
}

static void
signal_changed(GtkToggleButton *toggle, VMfmRecalcControls *controls)
{
    if (toggle && !gtk_toggle_button_get_active(toggle))
        return;

    controls->args->signal = gwy_radio_buttons_get_current(controls->signal);
    update_sensitivity(controls);
}

static const gchar base_amplitude_key[]  = "/module/vmfm_recalc/base_amplitude";
static const gchar base_frequency_key[]  = "/module/vmfm_recalc/base_frequency";
static const gchar quality_key[]         = "/module/vmfm_recalc/quality";
static const gchar signal_key[]          = "/module/vmfm_recalc/signal";
static const gchar spring_constant_key[] = "/module/vmfm_recalc/spring_constant";

static void
vmfm_recalc_sanitize_args(VMfmRecalcArgs *args)
{
    args->signal = CLAMP(args->signal, 0, SIGNAL_AMPLITUDE_M);
    args->spring_constant = CLAMP(args->spring_constant, 0.01, 1000);
    args->quality = CLAMP(args->quality, 0.01, 10000);
    args->base_frequency = CLAMP(args->base_frequency, 1, 1000000);
    args->base_amplitude = CLAMP(args->base_amplitude, 0.1, 100);
}

static void
vmfm_recalc_load_args(GwyContainer *container,
              VMfmRecalcArgs *args)
{
    *args = vmfm_recalc_defaults;

    gwy_container_gis_enum_by_name(container, signal_key, &args->signal);
    gwy_container_gis_double_by_name(container, spring_constant_key, &args->spring_constant);
    gwy_container_gis_double_by_name(container, quality_key, &args->quality);
    gwy_container_gis_double_by_name(container, base_frequency_key, &args->base_frequency);
    gwy_container_gis_double_by_name(container, base_amplitude_key, &args->base_amplitude);

    vmfm_recalc_sanitize_args(args);
}

static void
vmfm_recalc_save_args(GwyContainer *container,
              VMfmRecalcArgs *args)
{
    gwy_container_set_enum_by_name(container, signal_key, args->signal);
    gwy_container_set_double_by_name(container, spring_constant_key, args->spring_constant);
    gwy_container_set_double_by_name(container, quality_key, args->quality);
    gwy_container_set_double_by_name(container, base_frequency_key, args->base_frequency);
    gwy_container_set_double_by_name(container, base_amplitude_key, args->base_amplitude);

}


/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
