/*
 *  $Id: volume_linestat.c 20872 2018-03-20 22:11:52Z yeti-dn $
 *  Copyright (C) 2017-2018 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libprocess/brick.h>
#include <libprocess/arithmetic.h>
#include <libprocess/linestats.h>
#include <libprocess/gwyprocesstypes.h>
#include <libprocess/inttrans.h>
#include <libprocess/grains.h>
#include <libprocess/filters.h>
#include <libprocess/stats.h>
#include <libgwydgets/gwydataview.h>
#include <libgwydgets/gwylayer-basic.h>
#include <libgwydgets/gwyradiobuttons.h>
#include <libgwydgets/gwycombobox.h>
#include <libgwydgets/gwynullstore.h>
#include <libgwydgets/gwydgetutils.h>
#include <libgwymodule/gwymodule-volume.h>
#include <app/gwyapp.h>
#include <app/gwymoduleutils.h>

#define PSF_RUN_MODES (GWY_RUN_INTERACTIVE)

#define field_convolve_default(field, kernel) \
    gwy_data_field_area_ext_convolve((field), \
                                     0, 0, \
                                     gwy_data_field_get_xres(field), \
                                     gwy_data_field_get_yres(field), \
                                     (field), (kernel), \
                                     GWY_EXTERIOR_MIRROR_EXTEND, 0.0, TRUE)

enum {
    PREVIEW_SIZE = 360,
};

enum {
    RESPONSE_RESET   = 1,
    RESPONSE_ESTIMATE = 2,
};

typedef enum {
    GWY_PSF_DISPLAY_DATA   = 0,
    GWY_PSF_DISPLAY_PSF  = 1,
    GWY_PSF_DISPLAY_CONVOLVED = 2,
    GWY_PSF_DISPLAY_DIFFERENCE   = 3
} GwyPSFDisplayType;

typedef struct {
    GwyBrick *brick;
    GwyAppDataId op1;
    GwyAppDataId op2;
    gdouble sigma;
    gint zlevel;
    gint zlevelfit;
    GwyPSFDisplayType display;
} PSFArgs;

typedef struct {
    PSFArgs *args;
    GtkObject *sigma;
    GtkWidget *chooser_op2;
    GtkWidget *display;
    GtkObject *zlevel;
    GtkObject *zlevelfit;
    GwyContainer *mydata;
    GwyDataField *image;
    GtkWidget *dialog;
    GtkWidget *view;
    gboolean in_update;
    GwyDataField *resultview;
    GtkWidget *tf_width;
    GwySIValueFormat *vf;
} PSFControls;

static GwyAppDataId op2_id = GWY_APP_DATA_ID_NONE;

static gboolean module_register  (void);
static void     volume_psf       (GwyContainer *data,
                                  GwyRunType run);
static gboolean psf_dialog       (PSFArgs *args,
                                  GwyContainer *data);
static void     psf_do           (PSFArgs *args,
                                  GwyContainer *data);
static void     display_changed  (GtkComboBox *combo,
                                  PSFControls *controls);
static void     zlevel_changed   (PSFControls *controls,
                                  GtkAdjustment *adj);
static void     zlevelfit_changed(PSFControls *controls,
                                  GtkAdjustment *adj);
static void     sigma_changed    (GtkAdjustment *adj,
                                  PSFControls *controls);
static void     psf_sanitize_args(PSFArgs *args);
static void     psf_load_args    (GwyContainer *container,
                                  PSFArgs *args);
static void     psf_save_args    (GwyContainer *container,
                                  PSFArgs *args);
static gboolean psf_data_filter  (GwyContainer *data,
                                  gint id,
                                  gpointer user_data);
static void     psf_data_changed (GwyDataChooser *chooser,
                                  PSFControls *controls);
static void     preview          (PSFControls *controls,
                                  PSFArgs *args);
static void     estimate_sigma   (PSFArgs *args);
static gdouble  measure_tf_width (GwyDataField *psf);

static const PSFArgs psf_defaults = {
    NULL, GWY_APP_DATA_ID_NONE, GWY_APP_DATA_ID_NONE, 10,
    0, 0, GWY_PSF_DISPLAY_PSF,
};

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("Calculates the volume PSF."),
    "Petr Klapetek <pklapetek@gwyddion.net>",
    "1.0",
    "Petr Klapetek, Robb Puttock & David Nečas (Yeti)",
    "2018",
};

GWY_MODULE_QUERY2(module_info, volume_psf)

static gboolean
module_register(void)
{
    gwy_volume_func_register("volume_psf",
                             (GwyVolumeFunc)&volume_psf,
                             N_("/_Transfer Function Guess..."),
                             NULL,
                             PSF_RUN_MODES,
                             GWY_MENU_FLAG_VOLUME,
                             N_("Calculate volume transfer function via Wiener invert filter"));

    return TRUE;
}

static void
volume_psf(GwyContainer *data, GwyRunType run)
{
    PSFArgs args;
    GwyBrick *brick = NULL;

    g_return_if_fail(run & PSF_RUN_MODES);

    psf_load_args(gwy_app_settings_get(), &args);
    gwy_app_data_browser_get_current(GWY_APP_BRICK, &brick,
                                     GWY_APP_CONTAINER_ID, &args.op1.datano,
                                     GWY_APP_BRICK_ID, &args.op1.id,
                                     0);
    g_return_if_fail(GWY_IS_BRICK(brick));

    args.brick = brick;

    if (CLAMP(args.zlevel, 0, brick->zres-1) != args.zlevel)
        args.zlevel = 0;

    if (psf_dialog(&args, data))
        psf_do(&args, data);

    psf_save_args(gwy_app_settings_get(), &args);
}

static gboolean
psf_dialog(PSFArgs *args, GwyContainer *data)
{
    GtkWidget *dialog, *table, *hbox, *spin, *label;
    GwyDataChooser *chooser;
    PSFControls controls;
    GwyDataField *dfield;
    gint response, row;
    GwyPixmapLayer *layer;
    GwyBrick *brick;
    GQuark quark;
    const guchar *gradient;

    static const GwyEnum psf_outputs[] = {
        { N_("Data"),              GWY_PSF_DISPLAY_DATA,       },
        { N_("Transfer function"), GWY_PSF_DISPLAY_PSF,        },
        { N_("Convolved"),         GWY_PSF_DISPLAY_CONVOLVED,  },
        { N_("Difference"),        GWY_PSF_DISPLAY_DIFFERENCE, },
    };

    controls.args = args;
    brick = args->brick;

    controls.in_update = TRUE;

    dialog = gtk_dialog_new_with_buttons(_("Estimate Transfer Function"), NULL, 0,
                                         _("_Fit sigma"), RESPONSE_ESTIMATE,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

    controls.dialog = dialog;
    controls.resultview = gwy_data_field_new(1, 1, 1.0, 1.0, FALSE);
    dfield = controls.resultview;
    /* This sets pixel size, real dimensions, units, etc. */
    gwy_brick_extract_plane(brick, dfield, 0, 0, 0,
                            gwy_brick_get_xres(brick),
                            gwy_brick_get_yres(brick),
                            -1,
                            FALSE);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox,
                       FALSE, FALSE, 4);

    controls.mydata = gwy_container_new();
    gwy_container_set_object_by_name(controls.mydata, "/0/data", dfield);
    g_object_unref(dfield);

    quark = gwy_app_get_brick_palette_key_for_id(args->op1.id);
    if (gwy_container_gis_string(data, quark, &gradient)) {
        gwy_container_set_const_string_by_name(controls.mydata,
                                               "/0/base/palette", gradient);
    }
    controls.view = gwy_data_view_new(controls.mydata);
    layer = gwy_layer_basic_new();
    g_object_set(layer,
                 "data-key", "/0/data",
                 "gradient-key", "/0/base/palette",
                 NULL);
    gwy_data_view_set_data_prefix(GWY_DATA_VIEW(controls.view), "/0/data");
    gwy_data_view_set_base_layer(GWY_DATA_VIEW(controls.view), layer);
    gwy_set_data_preview_size(GWY_DATA_VIEW(controls.view), PREVIEW_SIZE);
    gtk_box_pack_start(GTK_BOX(hbox), controls.view, FALSE, FALSE, 0);

    table = gtk_table_new(2, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_table_set_col_spacings(GTK_TABLE(table), 6);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_box_pack_start(GTK_BOX(hbox), table, FALSE, FALSE, 4);
    row = 0;

    controls.chooser_op2 = gwy_data_chooser_new_channels();
    chooser = GWY_DATA_CHOOSER(controls.chooser_op2);
    gwy_data_chooser_set_active_id(chooser, &args->op2);
    gwy_data_chooser_set_filter(chooser, psf_data_filter, &args->op1, NULL);
    g_signal_connect(chooser, "changed",
                     G_CALLBACK(psf_data_changed), &controls);
    gwy_table_attach_adjbar(table, row, _("_Ideal response:"), NULL,
                            GTK_OBJECT(chooser), GWY_HSCALE_WIDGET_NO_EXPAND);
    gtk_table_set_row_spacing(GTK_TABLE(table), row, 8);
    row++;


    controls.display
        = gwy_enum_combo_box_new(psf_outputs, G_N_ELEMENTS(psf_outputs),
                                 G_CALLBACK(display_changed),
                                 &controls, args->display, TRUE);
    gwy_table_attach_adjbar(table, row, _("_Display:"), NULL,
                            GTK_OBJECT(controls.display),
                            GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;


    controls.zlevel = gtk_adjustment_new(args->zlevel, 0.0, brick->zres-1.0,
                                       1.0, 10.0, 0);
    gwy_table_attach_adjbar(table, row, _("_Z level for preview:"), _("px"),
                            controls.zlevel, GWY_HSCALE_LINEAR | GWY_HSCALE_SNAP);
    g_signal_connect_swapped(controls.zlevel, "value-changed",
                             G_CALLBACK(zlevel_changed), &controls);

    row++;

    controls.zlevelfit = gtk_adjustment_new(args->zlevelfit, 0.0, brick->zres-1.0,
                                       1.0, 10.0, 0);
    gwy_table_attach_adjbar(table, row, _("_Z level for sigma fit:"), _("px"),
                            controls.zlevelfit, GWY_HSCALE_LINEAR | GWY_HSCALE_SNAP);
    g_signal_connect_swapped(controls.zlevelfit, "value-changed",
                             G_CALLBACK(zlevelfit_changed), &controls);

    row++;


    controls.sigma = gtk_adjustment_new(args->sigma, 1e-5, 1e2, 1e-5, 1, 0);
    spin = gwy_table_attach_adjbar(table, row, _("_Sigma:"), "%",
                                   controls.sigma, GWY_HSCALE_LOG);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 7);
    g_object_set_data(G_OBJECT(controls.sigma), "controls", &controls);
    g_signal_connect(controls.sigma, "value-changed",
                     G_CALLBACK(sigma_changed), &controls);
    row++;

    controls.vf = gwy_data_field_get_value_format_xy(dfield,
                                                     GWY_SI_UNIT_FORMAT_VFMARKUP,
                                                     NULL);
    /* The width is usually rather small. */
    controls.vf->precision++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    label = gtk_label_new(_("TF width:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label,
                     0, 1, row, row+1, GTK_FILL, 0, 0, 0);

    label = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL(label), controls.vf->units);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label,
                     2, 3, row, row+1, GTK_FILL, 0, 0, 0);

    controls.tf_width = gtk_label_new(NULL);
    gtk_misc_set_alignment(GTK_MISC(controls.tf_width), 1.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), controls.tf_width,
                     1, 2, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    controls.in_update = FALSE;
    psf_data_changed(chooser, &controls);

    gtk_widget_show_all(dialog);
    do {
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            gtk_widget_destroy(dialog);
            case GTK_RESPONSE_NONE:
            gwy_si_unit_value_format_free(controls.vf);
            g_object_unref(controls.mydata);
            return FALSE;
            break;

            case GTK_RESPONSE_OK:
            break;

            case RESPONSE_ESTIMATE:
            gwy_app_wait_cursor_start(GTK_WINDOW(dialog));
            estimate_sigma(args);
            gwy_app_wait_cursor_finish(GTK_WINDOW(dialog));
            gtk_adjustment_set_value(GTK_ADJUSTMENT(controls.sigma),
                                     args->sigma);
            preview(&controls, args);
            break;


            default:
            g_assert_not_reached();
            break;
        }
    } while (response != GTK_RESPONSE_OK);

    gtk_widget_destroy(dialog);
    gwy_si_unit_value_format_free(controls.vf);
    g_object_unref(controls.mydata);

    return TRUE;
}

static void
sigma_changed(GtkAdjustment *adj, PSFControls *controls)
{
    controls->args->sigma = gtk_adjustment_get_value(adj);
    preview(controls, controls->args);
}

static void
display_changed(GtkComboBox *combo, PSFControls *controls)
{
    PSFArgs *args = controls->args;

    args->display = gwy_enum_combo_box_get_active(combo);
    preview(controls, controls->args);
}

static void
zlevel_changed(PSFControls *controls, GtkAdjustment *adj)
{
    controls->args->zlevel = gwy_adjustment_get_int(adj);
    preview(controls, controls->args);
}

static void
zlevelfit_changed(PSFControls *controls, GtkAdjustment *adj)
{
    controls->args->zlevelfit = gwy_adjustment_get_int(adj);
    preview(controls, controls->args);
}

static void
psf_data_changed(GwyDataChooser *chooser, PSFControls *controls)
{
    GwyAppDataId *object = &controls->args->op2;

    gwy_data_chooser_get_active_id(chooser, object);

    if (!controls->in_update)
        preview(controls, controls->args);
}

static gboolean
psf_data_filter(GwyContainer *data, gint id, gpointer user_data)
{
    GwyAppDataId *object = (GwyAppDataId*)user_data;
    GwyBrick *op1;
    GwyDataField *op2;
    GQuark quark;

    quark = gwy_app_get_data_key_for_id(id);
    op2 = GWY_DATA_FIELD(gwy_container_get_object(data, quark));

    data = gwy_app_data_browser_get(object->datano);

    quark = gwy_app_get_brick_key_for_id(object->id);
    op1 = GWY_BRICK(gwy_container_get_object(data, quark));

    return !gwy_data_field_check_compatibility_with_brick_xy
                                          (op2, op1,
                                           GWY_DATA_COMPATIBILITY_RES
                                           | GWY_DATA_COMPATIBILITY_REAL
                                           | GWY_DATA_COMPATIBILITY_LATERAL);
}

static void
estimate_sigma(PSFArgs *args)
{
    GwyDataField *measured, *ideal;
    GQuark quark;
    GwyContainer *mydata;
    GwyBrick *brick = args->brick;
    gint zlevel = args->zlevel;

    if (zlevel == -1)
        zlevel = 0;

    mydata = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    ideal = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    measured = gwy_data_field_new_alike(ideal, FALSE);
    gwy_brick_extract_plane(brick, measured, 0, 0, zlevel,
                            gwy_brick_get_xres(brick),
                            gwy_brick_get_yres(brick),
                            -1,
                            FALSE);

    args->sigma = gwy_data_field_find_regularization_sigma_for_psf(measured,
                                                                   ideal);
    /* Convert NSR to GUI parameter convention */
    args->sigma = 100.0*sqrt(args->sigma);

    g_object_unref(measured);
}

static void
preview(PSFControls *controls, PSFArgs *args)
{
    GwyDataField *dfield1, *dfield2, *psf, *convolved;
    GQuark quark;
    GwyContainer *mydata;
    GwyBrick *brick = args->brick;
    gint xres, yres, zlevel = args->zlevel;
    gdouble realsigma, tf_width;
    gchar *s;

    if (zlevel == -1)
        zlevel = 0;

    xres = gwy_brick_get_xres(brick);
    yres = gwy_brick_get_yres(brick);

    if (args->op2.datano < 0 || args->op2.id < 0) {
        if (args->display == GWY_PSF_DISPLAY_DATA) {
            gwy_brick_extract_plane(brick, controls->resultview, 0, 0, zlevel,
                                    xres, yres, -1, FALSE);
        }
        else
            gwy_data_field_clear(controls->resultview);
        gwy_data_field_data_changed(controls->resultview);
        return;
    }

    mydata = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    dfield2 = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    realsigma = args->sigma/100.0;
    dfield1 = gwy_data_field_new(1, 1, 1.0, 1.0, FALSE);
    gwy_brick_extract_plane(brick, dfield1, 0, 0,
                            zlevel, xres, yres, -1, FALSE);
    psf = gwy_data_field_new_alike(dfield1, TRUE);
    gwy_data_field_deconvolve_regularized(dfield1, dfield2, psf, realsigma);
    tf_width = measure_tf_width(psf);

    s = g_strdup_printf("%.*f",
                        controls->vf->precision,
                        tf_width/controls->vf->magnitude);
    gtk_label_set_text(GTK_LABEL(controls->tf_width), s);
    g_free(s);

    if (args->display == GWY_PSF_DISPLAY_DATA) {
        gwy_data_field_copy(dfield1, controls->resultview, TRUE);
        gwy_data_field_data_changed(controls->resultview);
    }
    else if (args->display == GWY_PSF_DISPLAY_PSF) {
         gwy_data_field_copy(psf, controls->resultview, TRUE);
         gwy_data_field_data_changed(controls->resultview);
    }
    else {
        convolved = gwy_data_field_new_alike(dfield1, TRUE);
        gwy_data_field_copy(dfield2, convolved, TRUE);
        field_convolve_default(convolved, psf);
        gwy_data_field_add(convolved, gwy_data_field_get_avg(dfield1));
        if (args->display == GWY_PSF_DISPLAY_DIFFERENCE)
            gwy_data_field_subtract_fields(convolved, convolved, dfield1);
        gwy_data_field_copy(convolved, controls->resultview, TRUE);
        gwy_data_field_data_changed(controls->resultview);
        g_object_unref(convolved);
    }
    g_object_unref(psf);
    g_object_unref(dfield1);
}

static void
psf_do(PSFArgs *args,
             GwyContainer *data)
{
    gint k, newid, xres, yres, zres;
    GQuark quarks[2], quark;
    GwyDataField *preview_dfield, *psf, *dfield1, *dfield2;
    GtkWidget *dialog;
    GwyContainer *mydata;
    gdouble realsigma;
    GtkWindow *window;
    GwyBrick *brick = args->brick;
    GwyBrick *result;
    const guchar *gradient;
    GwyGraphModel *gmodel;
    GwyGraphCurveModel *gcmodel;
    GwyDataLine *widths;

    if (args->op2.datano < 0 || args->op2.id < 0) {
        dialog = gtk_message_dialog_new(gwy_app_find_window_for_channel(data, args->op1.id),
                                        GTK_DIALOG_DESTROY_WITH_PARENT,
                                        GTK_MESSAGE_ERROR,
                                        GTK_BUTTONS_CLOSE,
                                        _("Wrong arguments for transfer function calculation."));
        gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        return;
    }

    window = gwy_app_find_window_for_volume(data, args->op1.id);
    gwy_app_wait_start(window, _("Calculating volume transfer function..."));

    xres = gwy_brick_get_xres(brick);
    yres = gwy_brick_get_yres(brick);
    zres = gwy_brick_get_zres(brick);

    dfield1 = gwy_data_field_new(xres, yres,
                                 gwy_brick_get_xreal(brick),
                                 gwy_brick_get_yreal(brick),
                                 FALSE);

    mydata = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    dfield2 = GWY_DATA_FIELD(gwy_container_get_object(mydata, quark));

    psf = gwy_data_field_new_alike(dfield1, TRUE);
    realsigma = args->sigma/100.0;

    result = gwy_brick_new_alike(brick, FALSE);
    widths = gwy_data_line_new(zres, gwy_brick_get_zreal(brick), FALSE);
    gwy_data_line_set_si_unit_x(widths, gwy_brick_get_si_unit_z(brick));

    for (k = 0; k < zres; k++) {
        gwy_brick_extract_plane(brick, dfield1, 0, 0, k, xres, yres, -1, FALSE);
        gwy_data_field_deconvolve_regularized(dfield1, dfield2, psf, realsigma);
        gwy_brick_set_plane(result, psf, 0, 0, k, xres, yres, -1);
        gwy_data_line_set_val(widths, k, measure_tf_width(psf));
        if (!gwy_app_wait_set_fraction(((gdouble)k)/(zres))) {
            g_object_unref(result);
            goto fail;
        }
    }
    gwy_data_line_set_si_unit_y(widths, gwy_data_field_get_si_unit_xy(psf));

    preview_dfield = gwy_data_field_new(xres, yres, xres, yres, FALSE);
    gwy_brick_mean_plane(result, preview_dfield, 0, 0, 0, xres, yres, -1, TRUE);

    newid = gwy_app_data_browser_add_brick(result, preview_dfield, data, TRUE);
    gwy_app_set_brick_title(data, newid, _("Volume TF"));
    gwy_app_volume_log_add_volume(data, args->op1.id, newid);

    quarks[0] = gwy_app_get_brick_palette_key_for_id(args->op1.id);
    quarks[1] = gwy_app_get_brick_palette_key_for_id(newid);
    if (gwy_container_gis_string(data, quarks[0], &gradient))
        gwy_container_set_const_string(data, quarks[1], gradient);

    gmodel = gwy_graph_model_new();
    gwy_graph_model_set_units_from_data_line(gmodel, widths);
    g_object_set(gmodel,
                 "title", _("TF width"),
                 "axis-label-left", _("TF width"),
                 "axis-label-bottom", _("z level"),
                 "si-unit-x", gwy_brick_get_si_unit_z(brick),
                 "si-unit-y", gwy_brick_get_si_unit_x(brick),
                 NULL);

    gcmodel = gwy_graph_curve_model_new();
    gwy_graph_curve_model_set_data_from_dataline(gcmodel, widths, -1, -1);
    g_object_set(gcmodel,
                 "description", _("TF width"),
                 "mode", GWY_GRAPH_CURVE_LINE,
                 NULL);
    gwy_graph_model_add_curve(gmodel, gcmodel);
    gwy_app_data_browser_add_graph_model(gmodel, data, TRUE);

fail:
    gwy_app_wait_finish();
    g_object_unref(widths);
    g_object_unref(psf);
    g_object_unref(dfield1);
}

static gdouble
measure_tf_width(GwyDataField *psf)
{
    GwyDataField *buf;
    gint xres, yres, xc, yc, i, j, k;
    const gdouble *d, *b;
    gdouble sx, sy, s2, sw, w, dx, dy, x, y;

    xres = gwy_data_field_get_xres(psf);
    yres = gwy_data_field_get_yres(psf);
    dx = gwy_data_field_get_xmeasure(psf);
    dy = gwy_data_field_get_ymeasure(psf);
    xc = xres/2;
    yc = yres/2;

    buf = gwy_data_field_duplicate(psf);
    gwy_data_field_threshold(buf, 0.05*gwy_data_field_get_max(buf), 0.0, 1.0);
    b = gwy_data_field_get_data_const(buf);
    if (b[yc*xres + xc] == 0.0) {
        g_warning("Cannot find the central TF peak.  Improve me!");
        g_object_unref(buf);
        return 0.0;
    }

    gwy_data_field_grains_extract_grain(buf, xc, yc);
    gwy_data_field_grains_grow(buf, 0.5*log(xres*yres),
                               GWY_DISTANCE_TRANSFORM_EUCLIDEAN, FALSE);
    d = gwy_data_field_get_data_const(psf);
    b = gwy_data_field_get_data_const(buf);
    sx = sy = sw = 0.0;
    k = 0;
    for (i = 0; i < yres; i++) {
        y = i*dy;
        for (j = 0; j < xres; j++, k++) {
            if (b[k] <= 0.0)
                continue;

            x = j*dx;
            w = fabs(d[k]);
            sw += w;
            sx += x*w;
            sy += y*w;
        }
    }
    sx /= sw;
    sy /= sw;
    s2 = 0.0;
    k = 0;
    for (i = 0; i < yres; i++) {
        y = i*dy;
        for (j = 0; j < xres; j++, k++) {
            if (b[k] <= 0.0)
                continue;

            x = j*dx;
            w = fabs(d[k]);
            s2 += ((x - sx)*(x - sx) + (y - sy)*(y - sy))*w;
        }
    }

    g_object_unref(buf);

    return sqrt(s2/sw);
}


static const gchar sigma_key[]     = "/module/volume_recalc/sigma";
static const gchar display_key[]   = "/module/volume_recalc/display";
static const gchar zlevel_key[]    = "/modules/volume_recalc/zlevel";
static const gchar zlevelfit_key[] = "/modules/volume_recalc/zlevelfit";

static void
psf_sanitize_args(PSFArgs *args)
{
    gwy_app_data_id_verify_channel(&args->op2);
}

static void
psf_load_args(GwyContainer *container, PSFArgs *args)
{
    *args = psf_defaults;

    gwy_container_gis_enum_by_name(container, display_key, &args->display);
    gwy_container_gis_double_by_name(container, sigma_key, &args->sigma);
    gwy_container_gis_int32_by_name(container, zlevel_key, &args->zlevel);
    gwy_container_gis_int32_by_name(container, zlevelfit_key, &args->zlevelfit);
    args->op2 = op2_id;

    psf_sanitize_args(args);
}

static void
psf_save_args(GwyContainer *container, PSFArgs *args)
{
    op2_id = args->op2;

    gwy_container_set_enum_by_name(container, display_key, args->display);
    gwy_container_set_double_by_name(container, sigma_key, args->sigma);
    gwy_container_set_int32_by_name(container, zlevel_key, args->zlevel);
    gwy_container_set_int32_by_name(container, zlevelfit_key, args->zlevelfit);
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
